;; Generated package description from diff-hl.el  -*- no-byte-compile: t -*-
(define-package "diff-hl" "1.9.0" "Highlight uncommitted changes using VC" '((cl-lib "0.2") (emacs "25.1")) :commit "37b00f3bad841e131d69442a89cbebc3041d996b" :authors '(("Dmitry Gutov" . "dgutov@yandex.ru")) :maintainer '("Dmitry Gutov" . "dgutov@yandex.ru") :keywords '("vc" "diff") :url "https://github.com/dgutov/diff-hl")
